from __future__ import division, print_function, absolute_import, unicode_literals

import sys
import six
from abc import ABCMeta, abstractmethod

if sys.version_info < (2, 7):
    from mog_commons.backported.inspect2 import getcallargs
else:
    from inspect import getcallargs

__all__ = [
    'String',
    'Unicode',
    'ListOf',
    'TupleOf',
    'SetOf',
    'DictOf',
    'VarArg',
    'KwArg',
    'types',
]

#
# Type definitions
#
String = six.string_types + (bytes,)

Unicode = unicode if six.PY2 else str


@six.add_metaclass(ABCMeta)
class ComposableType(object):
    """Label for composable types"""

    @abstractmethod
    def name(self):
        """abstract method"""

    @abstractmethod
    def check(self, obj):
        """abstract method"""


@six.add_metaclass(ABCMeta)
class IterableOf(ComposableType):
    def __init__(self, iterable_type, elem_type):
        self.iterable_type = iterable_type
        self.elem_type = elem_type

    def name(self):
        return '%s(%s)' % (self.iterable_type.__name__, _get_name(self.elem_type))

    def check(self, obj):
        return isinstance(obj, self.iterable_type) and all(_check_type(elem, self.elem_type) for elem in obj)


class ListOf(IterableOf):
    """Label for list element type assertion"""

    def __init__(self, elem_type):
        IterableOf.__init__(self, list, elem_type)


class TupleOf(IterableOf):
    """Label for tuple element type assertion"""

    def __init__(self, elem_type):
        IterableOf.__init__(self, tuple, elem_type)


class SetOf(IterableOf):
    """Label for set element type assertion"""

    def __init__(self, elem_type):
        IterableOf.__init__(self, set, elem_type)


class DictOf(ComposableType):
    """Label for dict element type assertion"""

    def __init__(self, key_type, value_type):
        self.key_type = key_type
        self.value_type = value_type

    def name(self):
        return 'dict(%s->%s)' % (_get_name(self.key_type), _get_name(self.value_type))

    def check(self, obj):
        return isinstance(obj, dict) and all(
            _check_type(k, self.key_type) and _check_type(v, self.value_type) for k, v in obj.items())


def VarArg(cls):
    """Shorthand description for var arg"""
    return TupleOf(cls)


def KwArg(cls):
    """Shorthand description for keyword arg"""
    return DictOf(String, cls)


#
# Helper functions
#
def _get_name(cls):
    if isinstance(cls, ComposableType):
        return cls.name()
    if isinstance(cls, tuple):
        return '(%s)' % '|'.join(_get_name(t) for t in cls)
    else:
        return cls.__name__


def _check_type(obj, cls):
    return cls.check(obj) if isinstance(cls, ComposableType) else isinstance(obj, cls)


#
# Decorators
#
def types(*return_type, **arg_types):
    """
    Assert types of the function arguments and return value.
    :param return_type: expected type of the return value
    :param arg_types: expected types of the arguments

    :example:
    @types(float, x=int, y=float, z=ListOf(int))
    def f(x, y, z):
        return x * y + sum(z)
    """
    assert len(return_type) <= 1, 'You can specify at most one return type.'

    arg_msg = '%s must be %s, not %s.'
    return_msg = 'must return %s, not %s.'

    def f(func):
        import functools

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # convert args to call args
            call_args = getcallargs(func, *args, **kwargs)

            for arg_name, expect in arg_types.items():
                assert arg_name in call_args, 'Not found argument: %s' % arg_name
                actual = call_args[arg_name]
                assert _check_type(actual, expect), arg_msg % (arg_name, _get_name(expect), type(actual).__name__)

            ret = func(*args, **kwargs)
            if return_type:
                assert _check_type(ret, return_type[0]), return_msg % (_get_name(return_type[0]), type(ret).__name__)
            return ret
        return wrapper

    return f



from __future__ import division, print_function, absolute_import, unicode_literals

import sys
import os
import errno
import subprocess
import six
from mog_commons.string import is_unicode
from mog_commons.functional import oget


#
# Process operations
#
def __convert_args(args, shell, cmd_encoding):
    # Note: workaround for http://bugs.python.org/issue8513
    workaround = shell and sys.version_info[:2] == (3, 2) and not sys.platform == 'win32'

    if isinstance(args, six.string_types):
        # input as string
        s = args.encode(cmd_encoding)
        if workaround:
            return ['/bin/sh', '-c', s], False
        else:
            return s, shell

    # input as list
    xs = ['/bin/sh', '-c'] if workaround else []
    if shell:
        args = [subprocess.list2cmdline(args)]
    for a in args:
        assert is_unicode(a), 'cmd must be unicode string, not %s' % type(a).__name__
        xs.append(a.encode(cmd_encoding))
    return xs, shell and not workaround


def __convert_env(env, encoding):
    """Environment variables should be bytes not unicode on Windows."""
    d = dict(os.environ, **(oget(env, {})))
    return dict((k.encode(encoding), v.encode(encoding)) for k, v in d.items())


def execute_command(args, shell=False, cwd=None, env=None, stdin=None, stdout=None, stderr=None, cmd_encoding='utf-8'):
    """
    Execute external command
    :param args: command line arguments : [unicode]
    :param shell: True when using shell : boolean
    :param cwd: working directory : string
    :param env: environment variables : dict
    :param stdin: standard input
    :param stdout: standard output
    :param stderr: standard error
    :param cmd_encoding: command line encoding: string
    :return: return code
    """
    args, shell = __convert_args(args, shell, cmd_encoding)
    return subprocess.call(args=args, shell=shell, cwd=cwd, env=__convert_env(env, cmd_encoding),
                           stdin=stdin, stdout=stdout, stderr=stderr)


def capture_command(args, shell=False, cwd=None, env=None, stdin=None, cmd_encoding='utf-8'):
    """
    Execute external command and capture output
    :param args: command line arguments : [string]
    :param shell: True when using shell : boolean
    :param cwd: working directory : string
    :param env: environment variables : dict
    :param stdin: standard input
    :param cmd_encoding: command line encoding: string
    :return: tuple of return code, stdout data and stderr data
    """
    args, shell = __convert_args(args, shell, cmd_encoding)
    p = subprocess.Popen(
        args, shell=shell, cwd=cwd, env=__convert_env(env, cmd_encoding),
        stdin=stdin, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout_data, stderr_data = p.communicate()
    return p.returncode, stdout_data, stderr_data


def execute_command_with_pid(args, pid_file=None, shell=False, cwd=None, env=None,
                             stdin=None, stdout=None, stderr=None, cmd_encoding='utf-8'):
    if pid_file is None:
        return execute_command(args, shell, cwd, env, stdin, stdout, stderr, cmd_encoding)
    else:
        try:
            args, shell = __convert_args(args, shell, cmd_encoding)
            p = subprocess.Popen(
                args, shell=shell, cwd=cwd, env=__convert_env(env, cmd_encoding),
                stdin=stdin, stdout=stdout, stderr=stderr)
            with open(pid_file, 'w') as f:
                f.write(str(p.pid))
            ret = p.wait()
        finally:
            # clean up pid file
            if pid_file is not None and os.path.exists(pid_file):
                os.remove(pid_file)
        return ret


def pid_exists(pid):
    # stole from https://github.com/giampaolo/psutil/blob/master/psutil/_psposix.py
    """Check whether pid exists in the current process table."""
    if pid == 0:
        # According to "man 2 kill" PID 0 has a special meaning:
        # it refers to <<every process in the process group of the
        # calling process>> so we don't want to go any further.
        # If we get here it means this UNIX platform *does* have
        # a process with id 0.
        return True
    try:
        os.kill(pid, 0)
    except OSError as err:
        if err.errno == errno.ESRCH:
            # ESRCH == No such process
            return False
        elif err.errno == errno.EPERM:
            # EPERM clearly means there's a process to deny access to
            return True
        else:
            # According to "man 2 kill" possible error values are
            # (EINVAL, EPERM, ESRCH) therefore we should never get
            # here. If we do let's be explicit in considering this
            # an error.
            raise err
    else:
        return True



from __future__ import division, print_function, absolute_import, unicode_literals

import six

__all__ = ['get_single_item', 'get_single_key', 'get_single_value', 'distinct']


def get_single_item(d):
    """Get an item from a dict which contains just one item."""
    assert len(d) == 1, 'Single-item dict must have just one item, not %d.' % len(d)
    return next(six.iteritems(d))


def get_single_key(d):
    """Get a key from a dict which contains just one item."""
    assert len(d) == 1, 'Single-item dict must have just one item, not %d.' % len(d)
    return next(six.iterkeys(d))


def get_single_value(d):
    """Get a value from a dict which contains just one item."""
    assert len(d) == 1, 'Single-item dict must have just one item, not %d.' % len(d)
    return next(six.itervalues(d))


def distinct(xs):
    """Get the list of distinct values with preserving order."""
    # don't use collections.OrderedDict because we do support Python 2.6
    seen = set()
    return [x for x in xs if x not in seen and not seen.add(x)]



from __future__ import division, print_function, absolute_import, unicode_literals

import sys
from mog_commons.string import is_unicode, to_bytes, to_unicode


def print_safe(str_or_bytes, encoding='utf-8', errors='ignore', output=sys.stdout, newline='\n'):
    """
    Print unicode or bytes universally.

    :param str_or_bytes: string
    :param encoding: encoding
    :param output: output file handler
    :param errors: error handling scheme. Refer to codecs.register_error.
    """
    writer = output.buffer if hasattr(output, 'buffer') else output

    # When the input type is bytes, verify it can be decoded with the specified encoding.
    decoded = str_or_bytes if is_unicode(str_or_bytes) else to_unicode(str_or_bytes, encoding, errors)
    encoded = to_bytes(decoded, encoding, errors)

    writer.write(encoded + to_bytes(newline, encoding, errors))
    output.flush()



from __future__ import division, print_function, absolute_import, unicode_literals


#
# functions for handling optional values
#
def omap(function, optional):
    """Map optional value"""
    return None if optional is None else function(optional)


def oget(optional, default=None):
    """Get optional value or default value"""
    return default if optional is None else optional


def ozip(*optionals):
    """Zip optional values. Return None if one value or the other is None."""
    return None if any(x is None for x in optionals) else tuple(optionals)



from setuptools import setup, find_packages

SRC_DIR = 'src'


def get_version():
    import sys

    sys.path[:0] = [SRC_DIR]
    return __import__('mog_commons').__version__


setup(
    name='mog-commons',
    version=get_version(),
    description='Common utility library for Python',
    author='mogproject',
    author_email='mogproj@gmail.com',
    license='Apache 2.0 License',
    url='https://github.com/mogproject/mog-commons-python',
    install_requires=[
        'six',
        'unittest2',
        'jinja2 == 2.6',  # specify library version to support Python 3.2
    ],
    tests_require=[
    ],
    package_dir={'': SRC_DIR},
    packages=find_packages(SRC_DIR),
    include_package_data=True,
    test_suite='tests',
)
